/*-------------------------------------------------
  QTMusicLib.c
  Routines to play standard midi messages via QT.
  Modified from QTMusic

Web site for this library and other ones I've done:
http://www.tunesmithy.co.uk/playmidilib/index.htm

E-mail support@tunesmithy.co.uk
  (c) Robert Walker 2002

For QTMusic
See http://www.mactech.com/articles/develop/issue_23/vanbrink.html
and 
http://developer.apple.com/samplecode/Sample_Code/QuickTime/Music_and_Sound/QTMusic.htm
  -------------------------------------------------*/
#include "QTMusicLib.h"
#include <QuickTimeMusic.h>
#include <Gestalt.h>
#include <QTML.h>

#define kInst_HammondOrgan     17
NoteChannel myNoteChannels_SME[MAX_QT_MIDI_CHANNELS*MAX_QT_GM_DEVICE_EMULATIONS];
int iPatchForGMChannel[MAX_QT_MIDI_CHANNELS*MAX_QT_GM_DEVICE_EMULATIONS];
NoteAllocator   myNoteAllocator_SME;
#include <windows.h>

#ifdef compile_QTM_as_DLL
int WINAPI DllMain (HINSTANCE hInstance, DWORD fdwReason, PVOID pvReserved)
{
 return TRUE ;
}
#endif

#define CHANN_FOR_IGM(iGmDevEm,iChannel) ((iChannel)+MAX_QT_MIDI_CHANNELS*(iGmDevEm))
extern QTM_DLL_EXPORT void QTM_DLL_CALLBACK  QTMusic_Delay (double dDelay)
{
 unsigned long   myDelay;
 Delay((int)(dDelay*60), &myDelay); 
}

ComponentResult QTMusic_GetNewChannel(int iChanForGM,NoteChannel *pmyNoteChannel)
{
 int iChannel=iChanForGM%MAX_QT_MIDI_CHANNELS;
 ComponentResult   myErr = noErr;
 NoteRequest    myNoteRequest;
 // fill out a note request, using NAStuffToneDescription
 myNoteRequest.info.flags = 0;
// myNoteRequest.info.reserved = 0;
 *(short *)(&myNoteRequest.info.polyphony) = EndianS16_NtoB(2);    // simultaneous tones
 *(Fixed *)(&myNoteRequest.info.typicalPolyphony) = EndianU32_NtoB(0x00010000);

	myErr = NAStuffToneDescription
         (myNoteAllocator_SME
         ,iChannel==NON_MELODIC_PERCUSSION_CHANNEL?kFirstDrumkit+1:max(1,iPatchForGMChannel[iChanForGM])
         // kInst_HammondOrgan
         , &myNoteRequest.tone
         );
 if (myErr != noErr)
  return myErr;
  
 // allocate a note channel
 myErr = NANewNoteChannel(myNoteAllocator_SME, &myNoteRequest, pmyNoteChannel);
 if (myErr != noErr)
  return myErr;
 /**doesn't work
 {
 NoteChannel        nNoteChanns=0;
  int seed=0;
  myErr=NAGetIndNoteChannel(myNoteAllocator_SME,0,&nNoteChanns,&seed);
  if (myErr != noErr)
  {
   return_value=FALSE;
   goto bail;
  }
 }
 **/
 myNoteChannels_SME[iChanForGM]=*pmyNoteChannel;
 return myErr;
}

extern QTM_DLL_EXPORT void QTM_DLL_CALLBACK  QTMusic_SendMidiEvent (int iStatus,int iChannel,int iData1,int iData2)
{
 QTMusic_SendMidiEventForGmDevEm (0,iStatus,iChannel,iData1,iData2);
}

extern QTM_DLL_EXPORT int QTM_DLL_CALLBACK  QTMusic_SendMidiEventForGmDevEm (int iGmDevEm,int iStatus,int iChannel,int iData1,int iData2)
{
 int return_value=TRUE;
 int iChanForGM=0;
 // here idea is to use the larger number of QT channels available to emulate a number
 // of identical GM devices each with the standard 16 midi channels
 NoteChannel    myNoteChannel= NULL;
 char found_chann=0;
 ComponentResult   myErr = noErr;
 iGmDevEm=max(0,min(iGmDevEm,MAX_QT_GM_DEVICE_EMULATIONS-1));
 iChanForGM=CHANN_FOR_IGM(iGmDevEm,iChannel);
 myNoteChannel = NULL;
 if(myNoteAllocator_SME==NULL)
 // open the note allocator component
 myNoteAllocator_SME = OpenDefaultComponent(kNoteAllocatorComponentType, 0);
 if (myNoteAllocator_SME == NULL)
 {
  return_value=FALSE;
  goto bail;
 }
 if(myNoteChannels_SME[iChanForGM])
  myNoteChannel=myNoteChannels_SME[iChanForGM];
 else
 myErr = QTMusic_GetNewChannel(iChanForGM,&myNoteChannel);
 if (myErr != noErr)
 {
  return_value=FALSE;
  goto bail;
 }
 switch(iStatus)
 {
  // http://developer.apple.com/techpubs/quicktime/qtdevdocs/RM/tp_rm_qtma.htm
 case MIDI_CONTROLLER:
  if(iChannel!=NON_MELODIC_PERCUSSION_CHANNEL)
   myErr=NASetController(myNoteAllocator_SME,myNoteChannel,iData1,iData2);
  // setting a controller in a channel seems to make it play melodic notes
  // - at least for Pan and All sound off, it did.
 case MIDI_SET_PATCH:
  myErr=NASetInstrumentNumber(myNoteAllocator_SME, myNoteChannel, iData1+1); 
  iPatchForGMChannel[iChanForGM]=iData1+1;
  break;
 case MIDI_NOTE_ON:
	 myErr=NAPlayNote(myNoteAllocator_SME, myNoteChannel, iData1,iData2);	
  break;
 case MIDI_NOTE_OFF:
  myErr=NAPlayNote(myNoteAllocator_SME, myNoteChannel, iData1,0); 
  break;
 case MIDI_PITCH_BEND:
  {
   int ipitch_bend=iData1+(iData2<<7);
   int ibend=(int)(0.5+(ipitch_bend-8192)*(double)0x100/(double)4096);
   if(ibend>=512)
    ibend=511;
   if(ibend<-512)
    ibend=-512;
   /**
   NADisposeNoteChannel(myNoteAllocator_SME, myNoteChannels_SME[iChanForGM]);
   myErr = QTMusic_GetNewChannel(iChanForGM,&myNoteChannel);
   if(myErr!=noErr)
   {
    return_value=FALSE;
    goto bail;
   }
   **/
   myErr=NASetController
    (myNoteAllocator_SME, myNoteChannel, kControllerPitchBend
    ,ibend
    );
   if(myErr!=noErr)
   {
    return_value=FALSE;
    goto bail;
   }
   break;
  }
 }
bail:;
 if(myErr!=noErr)
  return FALSE;
 return return_value;
}

extern QTM_DLL_EXPORT void QTM_DLL_CALLBACK QTMusic_DeallocateGmDevEm_SME (int iGmDevEm)
{
 int i=0;
 if(myNoteAllocator_SME)
 for(i=0;i<MAX_QT_MIDI_CHANNELS;i++)
  if(myNoteChannels_SME[CHANN_FOR_IGM(iGmDevEm,i)])
   NADisposeNoteChannel(myNoteAllocator_SME, myNoteChannels_SME[CHANN_FOR_IGM(iGmDevEm,i)]);
}

extern QTM_DLL_EXPORT void QTM_DLL_CALLBACK QTMusic_CleanUp_SME (void)
{
 int i=0;
 if(myNoteAllocator_SME)
 {
  for(i=0;i<MAX_QT_GM_DEVICE_EMULATIONS;i++)
   QTMusic_DeallocateGmDevEm_SME(i);
  CloseComponent(myNoteAllocator_SME);
 }
 ExitMovies();
 TerminateQTML();
}

extern QTM_DLL_EXPORT int QTM_DLL_CALLBACK  QTMusic_Initialise_SME (void) 
{
 long  myAttrs;
 OSErr   myErr = noErr;
 InitializeQTML(0L);
  
 // initialize QuickTime
 EnterMovies();

 myErr = Gestalt(gestaltQuickTime, &myAttrs);
 if (myErr == noErr)
 {
  return 1;
 }
 return 0;
}

